#!/usr/bin/env python3
"""A wrapper around the 'slicer' imaging tool (See also cli/_slicereport.py)"""

import logging
import subprocess
import sys
import csv
import json
import tempfile
from copy import copy
from pathlib import Path
from importlib.util import find_spec
if find_spec('bidscoin') is None:
    sys.path.append(str(Path(__file__).parents[2]))
from bidscoin import bcoin, bids, bidsversion, __version__

html_head = """<!DOCTYPE html>
<html lang="en">
<head>
    <title>Slice report</title>
    <style>
        body     { color: #D1D1D1; background-color: #181818; font-family: Arial; margin-left: 15px }
        h1       { color: Orange; font-size: 18px; display: inline-block; border: 1px solid orange;
                   padding: 10px 20px; border-radius: 10px;}
        a        { color: inherit; text-decoration: none; }
        a:hover  { color: Orange; }
        a:active { color: Yellow; }
    </style>
</head>

<body>

"""


def parse_options(options: list) -> str:
    """Check the OPTIONS arguments and return them to a string that can be passed to slicer"""
    for n, option in enumerate(options):
        if options[n-1] == 'l': continue                  # Skip checking the LUT string
        if not (option in ('L','l','i','e','t','n','u','s','c') or option.replace('.','').replace('-','').isdecimal()):
            print(f"Invalid OPTIONS: {' '.join(options)}"); sys.exit(2)
        if option.isalpha():
            options[n] = '-' + options[n]

    return  f"{' '.join(options)}"


def parse_outputs(outputargs: list, name: str) -> tuple:
    """
    Check the OUTPUT arguments and construct slicer and pngappend input arguments

    :return: (slicer_outputopts: str, pngappend_slices: str)
    """
    outputs  = ''       # These are input arguments for slicer (i.e. the slicer output images)
    slices   = ''       # These are input arguments for pngappend (i.e. the slicer output images)
    isnumber = lambda arg: arg.replace('.','').replace('-','').isdecimal()
    for n, outputarg in enumerate(outputargs):
        if not (outputarg in ('x', 'y', 'z', 'a', 'A', 'S', 'LF') or isnumber(outputarg)):
            print(f"Invalid {name}: '{outputarg}' in '{' '.join(outputargs)}'"); sys.exit(2)
        if outputarg.isalpha() and outputarg != 'LF':
            slices += f"{'' if n==0 else '-' if outputargs[n-1]=='LF' else '+'} slice_tmp{n}.png "
            if outputarg == 'a':
                outputs += f"-{outputarg} slice_tmp{n}.png "
            elif outputarg in ('x', 'y', 'z', 'A'):
                if not isnumber(outputargs[n+1]):
                    print(f"Invalid {name}: '{outputargs[n+1]}' in '{' '.join(outputargs)}'"); sys.exit(2)
                outputs += f"-{outputarg} {outputargs[n+1]} slice_tmp{n}.png "
            elif outputarg == 'S':
                if not (isnumber(outputargs[n+1]) and isnumber(outputargs[n+2])):
                    print(f"Invalid {name}: {outputarg} >> '{' '.join(outputargs)}'"); sys.exit(2)
                outputs += f"-{outputarg} {outputargs[n+1]} {outputargs[n+2]} slice_tmp{n}.png "

    return outputs, slices


def slicer_append(inputimage: Path, outlineimage: Path, mainopts: str, outputopts: str, sliceroutput: str, montage: Path, cluster: bool):
    """Run slicer and pngappend (locally or on the cluster) to create a montage of the sliced images"""

    # Create a workdir and the shell command
    workdir = Path(montage.parent)/next(tempfile._get_candidate_names())
    workdir.mkdir()
    command = f"cd {workdir}\n" \
              f"slicer {inputimage} {outlineimage} {mainopts} {outputopts}\n" \
              f"pngappend {sliceroutput} {montage.name}\n" \
              f"mv {montage.name} {montage.parent}\n" \
              f"rm -r {workdir}"

    # Wrap the command
    if cluster:
        if inputimage.stat().st_size > 50 * 1024**2:
            mem = '8gb'         # Ask for more resources if we have a large (e.g. 4D) input image
        else:
            mem = '1gb'
        command = f"qsub -l walltime=0:02:00,mem={mem} -N slicereport -e {tempfile.gettempdir()} -o {tempfile.gettempdir()}<<EOF\n{command}\nEOF"

    # Run the command
    LOGGER.bcdebug(f"Command: {command}")
    process = subprocess.run(command, shell=True, capture_output=True, text=True)
    if process.stderr or process.returncode != 0:
        LOGGER.warning(f"{command}\nErrorcode {process.returncode}:\n{process.stdout}\n{process.stderr}")
    if process.returncode != 0:
        sys.exit(process.returncode)


def slicereport(bidsdir: str, pattern: str, outlinepattern: str, outlineimage: str, subjects: list, reportdir: str, crossdirs: str, qccols: list, cluster: bool, options: list, outputs: list, suboptions: list, suboutputs: list):
    """
    :param bidsdir:         The bids-directory with the subject data
    :param pattern:         Globlike search pattern to select the images in bidsdir to be reported, e.g. 'anat/*_T1w*'
    :param outlinepattern:  Globlike search pattern to select red-outline images that are projected on top of the reported images. Prepend `outlinedir:` if your outline images are in `outlinedir` instead of `bidsdir`
    :param outlineimage:    A common red-outline image that is projected on top of all images
    :param subjects:        Space separated list of sub-# identifiers to be processed (the sub-prefix can be left out). If not specified then all sub-folders in the bidsfolder will be processed
    :param reportdir:       The folder where the report is saved
    :param crossdirs:       A (list of) folder(s) with cross-linked sub-reports
    :param qccols:          Column names for creating an accompanying tsv-file to store QC-rating scores
    :param cluster:         Use qsub to submit the slicer jobs to a high-performance compute (HPC) cluster
    :param options:         Slicer main options
    :param outputs:         Slicer output options
    :param suboptions:      Slicer main options for creating the sub-reports (same as OPTIONS)
    :param suboutputs:      Slicer output options for creating the sub-reports (same as OUTPUTS)
    :return:
    """

    # Input checking
    bidsdir = Path(bidsdir).resolve()
    if not bidsdir.is_dir():
        print(f"Could not find the bids folder: {bidsdir}"); return
    if not reportdir:
        reportdir = bidsdir/'derivatives/slicereport'
    else:
        reportdir = Path(reportdir).resolve()
    if isinstance(crossdirs, str):
        crossdirs = [crossdirs]
    elif crossdirs is None:
        crossdirs = []
    for crossdir in crossdirs:
        if not Path(crossdir).is_dir():
            print(f"Could not find: {crossdir}"); return
    if outlineimage:
        if outlinepattern:
            print('The "--outlineimage" and "--outlinepattern" arguments are mutually exclusive, please specify one or the other'); return
        outlineimage = Path(outlineimage).resolve()
        if not outlineimage.is_file():
            print(f"Could not find the common outline image: {outlineimage}"); return
    if outlinepattern and ':' in outlinepattern:
        outlinedir, outlinepattern = outlinepattern.split(':',1)
        outlinedir = Path(outlinedir).resolve()
        if not outlinedir.is_dir():
            print(f"Could not find the outline folder: {outlinedir}"); return
    else:
        outlinedir = bidsdir
    if not suboptions:
        suboptions = copy(options)

    # Format the slicer main and output options and sliced images
    options                     = parse_options(options)
    outputs, sliceroutput       = parse_outputs(outputs, 'OUTPUTS')
    suboptions                  = parse_options(suboptions)
    suboutputs, subsliceroutput = parse_outputs(suboutputs, 'SUBOUTPUTS')

    # Get the list of subjects
    if not subjects:
        subjects = bcoin.lsdirs(bidsdir, 'sub-*')
        if not subjects:
            print(f"No subjects found in: {bidsdir/'sub-*'}"); return
    else:
        subjects = ['sub-' + subject.replace('sub-', '') for subject in subjects]               # Make sure there is a "sub-" prefix
        subjects = [bidsdir/subject for subject in subjects if (bidsdir/subject).is_dir()]

    # Start logging
    reportdir.mkdir(parents=True, exist_ok=True)
    (reportdir/'slicereport.log').unlink(missing_ok=True)
    bcoin.setup_logging(reportdir/'slicereport.log')
    LOGGER.info(f"Command: slicereport {' '.join(sys.argv[1:])}")

    # Create the report index file
    report = reportdir/'index.html'
    report.write_text(f'{html_head}<h1>Command:<span style="color: White"> slicereport {" ".join(sys.argv[1:])}</span></h1>\n')

    # Create a QC tsv-file
    qcfile = reportdir/'qcscores.tsv'
    if qccols:
        with open(qcfile, 'wt') as fid:
            tsv_writer = csv.writer(fid, delimiter='\t')
            tsv_writer.writerow(['subject/session'] + qccols)

    # Loop over the subject/session-directories
    for subject in subjects:
        sessions = bcoin.lsdirs(subject, 'ses-*')
        if not sessions:
            sessions = [subject]
        for session in sessions:

            # Write a row in the QC tsv-file
            if qccols:
                with open(qcfile, 'a') as fid:
                    tsv_writer = csv.writer(fid, delimiter='\t')
                    tsv_writer.writerow([str(session.relative_to(bidsdir))] + len(qccols) * ['n/a'])

            # Search for the image(s) to report
            LOGGER.info(f"Processing images in: {session.relative_to(bidsdir)}")
            images = sorted([match for match in session.glob(pattern) if '.nii' in match.suffixes])
            if not images:
                LOGGER.warning(f"Could not find images using: {session.relative_to(bidsdir)}/{pattern}")
                continue
            outlineimages = [''] * len(images)
            if outlinepattern:
                outlinesession = outlinedir/session.relative_to(bidsdir)
                outlineimages  = sorted([match.with_suffix('').with_suffix('') for match in outlinesession.glob(outlinepattern) if '.nii' in match.suffixes])
                if len(outlineimages) != len(images):
                    LOGGER.error(f"Nr of outline images ({len(outlineimages)}) in {outlinesession.relative_to(bidsdir)} should be the same as the number of underlying images ({len(images)})")
                    outlineimages = [''] * len(images)

            # Generate a report row and a sub-report for each session
            reportses = reportdir/session.relative_to(bidsdir)
            reportses.mkdir(parents=True, exist_ok=True)
            for n, image in enumerate(images):

                # Generate the sliced image montage
                outline = outlineimages[n] if outlinepattern else outlineimage
                montage = reportses/image.with_suffix('').with_suffix('.png').name
                slicer_append(image, outline, options, outputs, sliceroutput, montage, cluster)

                # Add the montage as a (sub-report linked) row to the report
                caption   = f"{image.relative_to(bidsdir)}{'&nbsp;&nbsp;&nbsp;( ../'+str(outline.relative_to(outlinesession))+' )' if outlinepattern and outline else ''}"
                subreport = reportses/f"{bids.insert_bidskeyval(image, 'desc', 'subreport', False).with_suffix('').stem}.html"
                with report.open('a') as fid:
                    fid.write(f'\n<p><a href="{subreport.relative_to(reportdir).as_posix()}"><image src="{montage.relative_to(reportdir).as_posix()}"><br>\n{caption}</a></p>\n')

                # Add the sub-report
                if suboutputs:
                    montage = subreport.with_suffix('.png')
                    slicer_append(image, outline, suboptions, suboutputs, subsliceroutput, montage, cluster)
                crossreports = ''
                for crossdir in crossdirs:          # Include niprep reports
                    for crossreport in sorted(Path(crossdir).glob(f"{subject.name.split('_')[0]}*.html")) + sorted((Path(crossdir)/session.relative_to(bidsdir)).glob('*.html')):
                        crossreports += f'\n<br><a href="{crossreport.resolve()}">&#8618; {crossreport}</a>'
                if subreport.with_suffix('.json').is_file():
                    with open(subreport.with_suffix('.json'), 'r') as meta_fid:
                        metadata = f"\n\n<p>{json.load(meta_fid)}</p>"
                elif image.with_suffix('').with_suffix('.json').is_file():
                    with open(image.with_suffix('').with_suffix('.json'), 'r') as meta_fid:
                        metadata = f"\n\n<p>{json.load(meta_fid)}</p>"
                else:
                    metadata = ''
                subreport.write_text(f'{html_head}<h1>{caption}</h1>\n{crossreports}\n<p><image src="{montage.name}"></p>{metadata}\n\n</body></html>')

    # Create a dataset description file if it does not exist
    dataset = reportdir/'dataset_description.json'
    if not dataset.is_file():
        description = {"Name": "Slicereport - A visual inspection report",
                       "BIDSVersion": bidsversion(),
                       "DatasetType": "derivative",
                       "GeneratedBy": [{"Name":"BIDScoin", "Version":__version__, "CodeURL":"https://github.com/Donders-Institute/bidscoin"}]}
        with dataset.open('w') as fid:
            json.dump(description, fid, indent=4)

    # Finish off
    errors = bcoin.reporterrors().replace('\n', '<br>\n')
    if errors:
        footer = '<h3 style="color: Red">The following errors and warnings were reported:</h3>\n'
    else:
        footer = '<h3 style="color: LimeGreen">No errors or warnings were reported</h3>\n'
    with report.open('a') as fid:
        fid.write(f'\n<br>{footer}<p>\n{errors}</p>\n\n</body></html>')
    if qccols:
        LOGGER.info(' ')
        LOGGER.info('To store QC ratings, open:')
        LOGGER.info(qcfile)
    LOGGER.info(' ')
    LOGGER.info('To view the slice report, point your web browser at:')
    LOGGER.info(f"{report}\n ")
    if cluster:
        LOGGER.info(f"But first wait for your `slicereport`-jobs to finish... Use e.g.:\n\nqstat $(qselect -s RQ) | grep slicereport\n\n(Standard error/output files are written to: {tempfile.gettempdir()})\n")


def main():
    """Console script entry point"""

    from bidscoin.cli._slicereport import get_parser

    args = get_parser().parse_args()
    slicereport(bidsdir        = args.bidsfolder,
                pattern        = args.pattern,
                outlinepattern = args.outlinepattern,
                outlineimage   = args.outlineimage,
                subjects       = args.participant_label,
                reportdir      = args.reportfolder,
                crossdirs      = args.xlinkfolder,
                qccols         = args.qcscores,
                cluster        = args.cluster,
                options        = args.options,
                outputs        = args.outputs,
                suboptions     = args.suboptions,
                suboutputs     = args.suboutputs)


if __name__ == '__main__':
    LOGGER = logging.getLogger(f"bidscoin.{Path(__file__).stem}")
    main()

else:
    LOGGER = logging.getLogger(__name__)



import copy
import pandas as pd
import pytest
import shutil
import re
import json
import ruamel.yaml.comments
from pathlib import Path
from nibabel.testing import data_path
from pydicom.data import get_testdata_file
from bidscoin import bcoin, bids, bidsmap_template

bcoin.setup_logging()


@pytest.fixture(scope='module')
def dcm_file():
    return Path(get_testdata_file('MR_small.dcm'))


@pytest.fixture(scope='module')
def dicomdir():
    return Path(get_testdata_file('DICOMDIR'))


@pytest.fixture(scope='module')
def par_file():
    return Path(data_path)/'phantom_EPI_asc_CLEAR_2_1.PAR'


@pytest.fixture(scope='module')
def test_bidsmap():
    """The path to the study bidsmap `test_data/bidsmap.yaml`"""
    return Path(__file__).parent/'test_data'/'bidsmap.yaml'


class TestDataSource:
    """Test the bids.DataSource class"""

    @pytest.fixture()
    def datasource(self, dcm_file):
        return bids.DataSource(dcm_file, {'dcm2niix2bids': {}}, 'DICOM')

    @pytest.fixture()
    def extdatasource(self, dcm_file, tmp_path):
        ext_dcm_file = shutil.copyfile(dcm_file, tmp_path/dcm_file.name)
        with ext_dcm_file.with_suffix('.json').open('w') as sidecar:
            json.dump({'PatientName': 'ExtendedAttributesTest'}, sidecar)
        return bids.DataSource(ext_dcm_file, {'dcm2niix2bids': {}}, 'DICOM')

    def test_is_datasource(self, datasource):
        assert datasource.is_datasource()
        assert datasource.dataformat == 'DICOM'

    def test_properties(self, datasource):
        assert datasource.properties( 'filepath:.*/(.*?)_files/.*') == 'test'   # path = [..]/pydicom/data/test_files/MR_small.dcm'
        assert datasource.properties(r'filename:MR_(.*?)\.dcm')     == 'small'
        assert datasource.properties( 'filesize')                   == '9.60 kB'
        assert datasource.properties( 'nrfiles')                    == 76

    def test_attributes(self, datasource, extdatasource):
        assert datasource.attributes(r'PatientName:.*\^(.*?)1') == 'MR'         # PatientName = 'CompressedSamples^MR1'
        assert extdatasource.attributes('PatientName')          == 'ExtendedAttributesTest'

    @pytest.mark.parametrize('subid',  ['sub-001', 'pat^visit'])
    @pytest.mark.parametrize('sesid',  ['ses-01',  'visit^01', ''])
    @pytest.mark.parametrize('subprefix', ['sub-', 'pat^', '*'])
    @pytest.mark.parametrize('sesprefix', ['ses-', 'visit^', '*'])
    def test_subid_sesid(self, subid, sesid, subprefix, sesprefix, tmp_path, dcm_file):
        subsesdir     = tmp_path/'data'/subid/sesid
        subsesdir.mkdir(parents=True)
        subses_file   = shutil.copy(dcm_file, subsesdir)
        subses_source = bids.DataSource(subses_file, {'dcm2niix2bids': {}}, 'DICOM', subprefix=subprefix, sesprefix=sesprefix)
        sub, ses      = subses_source.subid_sesid(f"<<filepath:/data/{subses_source.resubprefix()}(.*?)/>>", f"<<filepath:/data/{subses_source.resubprefix()}.*?/{subses_source.resesprefix()}(.*?)/>>")
        expected_sub  = 'sub-' + bids.cleanup_value(re.sub(f"^{subses_source.resubprefix()}", '', subid)  if subid.startswith(subprefix)  or subprefix=='*' else '')  # NB: this expression is too complicated / resembles the actual code too much :-/
        expected_ses  = 'ses-' + bids.cleanup_value(re.sub(f"^{subses_source.resesprefix()}", '', sesid)) if (subid.startswith(subprefix) or subprefix=='*') and (sesid.startswith(sesprefix) or sesprefix=='*') and sesid else ''
        print(f"[{subprefix}, {subid}] -> {sub}\t\t[{sesprefix}, {sesid}] -> {ses}")
        assert (sub, ses) == (expected_sub, expected_ses)
        assert subses_source.subid_sesid(r'<<PatientName:.*\^(.*?)1>>', '') == ('sub-MR', '')

    def test_dynamicvalue(self, datasource):
        assert datasource.dynamicvalue(r'<PatientName>')                                         == 'CompressedSamplesMR1'
        assert datasource.dynamicvalue(r'PatientName:.*\^(.*?)1')                                == r'PatientName:.*\^(.*?)1'
        assert datasource.dynamicvalue(r'<PatientName:.*\^(.*?)1>')                              == 'MR'
        assert datasource.dynamicvalue(r'<<PatientName:.*\^(.*?)1>>')                            == r'<<PatientName:.*\^(.*?)1>>'
        assert datasource.dynamicvalue(r'<<PatientName:.*\^(.*?)1>>', runtime=True)              == 'MR'
        assert datasource.dynamicvalue(r'pat-<PatientName:.*\^(.*?)1>I<filename:MR_(.*?)\.dcm>') == 'patMRIsmall'
        assert datasource.dynamicvalue(r"<Patient's Name>")                                      == 'CompressedSamplesMR1'    # Patient's Name, 0x00100010, 0x10,0x10, (0x10, 0x10), and (0010, 0010) index keys are all equivalent
        assert datasource.dynamicvalue(r'<0x00100010>')                                          == 'CompressedSamplesMR1'
        assert datasource.dynamicvalue(r'<0x10,0x10>')                                           == 'CompressedSamplesMR1'
        assert datasource.dynamicvalue(r'<(0x10, 0x10)>')                                        == 'CompressedSamplesMR1'
        assert datasource.dynamicvalue(r'<(0010, 0010)>')                                        == 'CompressedSamplesMR1'

def test_unpack(dicomdir, tmp_path):
    unpacked = bids.unpack(dicomdir.parent, '', tmp_path)
    assert unpacked[1]
    assert len(unpacked[0]) == 6
    for ses in unpacked[0]:
        assert 'Doe^Archibald' in ses.parts or 'Doe^Peter' in ses.parts


def test_is_dicomfile(dcm_file):
    assert bids.is_dicomfile(dcm_file)


def test_is_parfile(par_file):
    assert bids.is_parfile(par_file)


def test_get_dicomfile(dcm_file, dicomdir):
    assert bids.get_dicomfile(dcm_file.parent).name == '693_J2KI.dcm'
    assert bids.get_dicomfile(dicomdir.parent).name == '6154'


def test_get_datasource(dicomdir):
    datasource = bids.get_datasource(dicomdir.parent, {'dcm2niix2bids': {}})
    assert datasource.is_datasource()
    assert datasource.dataformat == 'DICOM'


@pytest.mark.parametrize('template', bcoin.list_plugins()[1])
def test_load_check_template(template):

    # Load a valid template
    bidsmap, _ = bids.load_bidsmap(template, checks=(False, False, False))
    assert isinstance(bidsmap, dict) and bidsmap
    assert bids.check_template(bidsmap) == True

    # Add an invalid datatype
    bidsmap['DICOM']['foo'] = bidsmap['DICOM']['extra_data']
    assert bids.check_template(bidsmap) == False
    del bidsmap['DICOM']['foo']

    # Remove a valid suffix (BIDS-entity)
    bidsmap['DICOM']['anat'].pop(-2)        # NB: Assumes CT is the last item, MTR the second last
    assert bids.check_template(bidsmap) == False


def test_match_runvalue():
    assert bids.match_runvalue('my_pulse_sequence_name', '_name')      == False
    assert bids.match_runvalue('my_pulse_sequence_name', '^my.*name$') == True
    assert bids.match_runvalue('T1_MPRage', '(?i).*(MPRAGE|T1w).*')    == True
    assert bids.match_runvalue('', None)                               == True
    assert bids.match_runvalue(None, '')                               == True
    assert bids.match_runvalue(  [1, 2, 3],     [1,2,  3])             == True
    assert bids.match_runvalue(  [1,2,  3],    '[1, 2, 3]')            == True
    assert bids.match_runvalue(  [1, 2, 3],  r'\[1, 2, 3\]')           == True
    assert bids.match_runvalue( '[1, 2, 3]',   '[1, 2, 3]')            == True
    assert bids.match_runvalue( '[1, 2, 3]', r'\[1, 2, 3\]')           == True
    assert bids.match_runvalue( '[1, 2, 3]',    [1, 2, 3])             == True
    assert bids.match_runvalue( '[1,2,  3]',    [1,2,  3])             == False
    assert bids.match_runvalue(r'\[1, 2, 3\]',  [1, 2, 3])             == False


def test_load_bidsmap(test_bidsmap):

    # Test loading with recommended arguments for load_bidsmap
    full_arguments_map, return_path = bids.load_bidsmap(Path(test_bidsmap.name), test_bidsmap.parent)
    assert type(full_arguments_map) == ruamel.yaml.comments.CommentedMap
    assert full_arguments_map is not []

    # Test loading with no input folder0, should load default from heuristics folder
    no_input_folder_map, _ = bids.load_bidsmap(test_bidsmap)
    assert type(no_input_folder_map) == ruamel.yaml.comments.CommentedMap
    assert no_input_folder_map is not []

    # Test loading with full path to only bidsmap file
    full_path_to_bidsmap_map, _ = bids.load_bidsmap(test_bidsmap)
    assert type(full_path_to_bidsmap_map) == ruamel.yaml.comments.CommentedMap
    assert no_input_folder_map is not []


def test_validate_bidsmap(test_bidsmap):

    # Load a BIDS-valid study bidsmap
    bidsmap, _ = bids.load_bidsmap(test_bidsmap)
    run        = bidsmap['DICOM']['func'][0]
    assert bids.validate_bidsmap(bidsmap) == True

    # Validate the bids-keys
    run['bids']['flip'] = 'foo'     # Add a false key
    assert bids.validate_bidsmap(bidsmap) == False
    del run['bids']['flip']
    del run['bids']['task']         # Remove a required key
    assert bids.validate_bidsmap(bidsmap) == False
    run['bids']['task'] = 'foo'

    # Check bids-suffix
    run['bids']['suffix'] = 'T1w'   # Set an invalid suffix
    assert bids.validate_bidsmap(bidsmap) == False
    run['bids']['suffix'] = 'bold'

    # Check bids-values
    run['bids']['task'] = ''        # Remove a required value
    assert bids.validate_bidsmap(bidsmap) == False
    run['bids']['task'] = 'f##'     # Add invalid characters (they are cleaned out)
    assert bids.validate_bidsmap(bidsmap) == True
    run['bids']['task'] = 'foo'
    run['bids']['run']  = 'a'       # Add an invalid (non-numeric) index
    assert bids.validate_bidsmap(bidsmap) == False


def test_check_bidsmap(test_bidsmap):

    # Load a template and a study bidsmap
    template_bidsmap, _ = bids.load_bidsmap(bidsmap_template, checks=(True, True, False))
    study_bidsmap, _    = bids.load_bidsmap(test_bidsmap)

    # Test the output of the template bidsmap
    checks   = (True, True, False)
    is_valid = bids.check_bidsmap(template_bidsmap, checks)
    for each, check in zip(is_valid, checks):
        assert each in (None, True, False)
        if check:
            assert each in (None, True)

    # Test the output of the study bidsmap
    is_valid = bids.check_bidsmap(study_bidsmap, checks)
    for each, check in zip(is_valid, checks):
        assert each in (None, True, False)
        if check:
            assert each == True


def test_check_run(test_bidsmap):

    # Load a bidsmap
    bidsmap, _ = bids.load_bidsmap(test_bidsmap)

    # Collect the first func run-item
    checks = (True, True, True)             # = (keys, suffixes, values)
    run    = bidsmap['DICOM']['func'][0]

    # Check datatype
    assert bids.check_run('func', run, checks) == (True, True, True)
    assert bids.check_run('anat', run, checks) == (None, False, None)
    assert bids.check_run('foo',  run, checks) == (None, None, None)

    # Check bids-keys
    run['bids']['flip'] = 'foo'     # Add a false key
    assert bids.check_run('func', run, checks) == (False, True, None)
    del run['bids']['flip']
    del run['bids']['acq']          # Remove a valid key
    assert bids.check_run('func', run, checks) == (False, True, True)
    run['bids']['acq'] = 'foo'

    # Check bids-suffix
    run['bids']['suffix'] = 'T1w'   # Set an invalid suffix
    assert bids.check_run('func', run, checks) == (None, False, None)
    run['bids']['suffix'] = 'bold'

    # Check bids-values
    run['bids']['task'] = ''        # Remove a required value
    assert bids.check_run('func', run, checks) == (True, True, False)
    run['bids']['task'] = 'f##'     # Add invalid characters
    assert bids.check_run('func', run, checks) == (True, True, False)
    run['bids']['task'] = 'foo'
    run['bids']['run']  = 'a'       # Add an invalid (non-numeric) index
    assert bids.check_run('func', run, checks) == (True, True, False)


def test_check_ignore():

    bidsignore = 'mrs/;sub-*_foo.*;*foo/sub-*_bar.*'

    assert bids.check_ignore('mrs',                bidsignore)         == True      # Test default: datatype = 'dir'
    assert bids.check_ignore('mrs',                bidsignore, 'file') == False
    assert bids.check_ignore('mrs/sub-01_foo.nii', bidsignore, 'dir')  == False
    assert bids.check_ignore('mrs/sub-01_bar.nii', bidsignore, 'file') == False
    assert bids.check_ignore('foo/sub-01_bar.nii', bidsignore, 'file') == True
    assert bids.check_ignore('bar/sub-01_bar.nii', bidsignore, 'file') == False
    assert bids.check_ignore('bar/sub-01_foo.nii', bidsignore, 'file') == False
    assert bids.check_ignore('sub-01_foo.nii',     bidsignore, 'file') == True


def test_find_run(test_bidsmap):

    # Load a bidsmap and create a duplicate dataformat section
    bidsmap, _     = bids.load_bidsmap(test_bidsmap)
    bidsmap['PET'] = copy.deepcopy(bidsmap['DICOM'])

    # Collect provenance of the first anat run-item
    provenance                              = bidsmap['DICOM']['anat'][0]['provenance']
    tag                                     = '123456789'
    bidsmap['PET']['anat'][0]['provenance'] = tag

    # Find run with the wrong dataformat
    run = bids.find_run(bidsmap, provenance, dataformat='PET')
    assert run == {}

    # Find run with the wrong datatype
    run = bids.find_run(bidsmap, provenance, datatype='func')
    assert run == {}

    # Find run with partial provenance
    run = bids.find_run(bidsmap, 'sub-001')
    assert run == {}

    # Find run with full provenance
    run = bids.find_run(bidsmap, provenance)
    assert isinstance(run, dict)
    run = bids.find_run(bidsmap, tag, dataformat='PET', datatype='anat')
    assert run.get('provenance') == tag


def test_delete_run(test_bidsmap):

    # Load a study bidsmap and delete one anat run
    bidsmap, _ = bids.load_bidsmap(test_bidsmap)
    nritems    = len(bidsmap['DICOM']['anat'])
    provenance = bidsmap['DICOM']['anat'][0]['provenance']
    bids.delete_run(bidsmap, provenance)

    assert len(bidsmap['DICOM']['anat']) == nritems - 1
    assert bids.find_run(bidsmap, provenance) == {}


def test_append_run(test_bidsmap):

    # Load a study bidsmap and delete one anat run
    bidsmap, _ = bids.load_bidsmap(test_bidsmap)

    # Collect and modify the first anat run-item
    run                          = copy.deepcopy(bidsmap['DICOM']['anat'][0])
    run['datasource'].dataformat = 'Foo'
    run['datasource'].datatype   = 'Bar'

    # Append the run elsewhere in the bidsmap
    bids.append_run(bidsmap, run)
    assert bidsmap['Foo']['Bar'][0]['provenance'] == run['provenance']


def test_update_bidsmap(test_bidsmap):

    # Load a study bidsmap and move the first run-item from func to anat
    bidsmap, _ = bids.load_bidsmap(test_bidsmap)

    # Collect and modify the first func run-item
    run                        = copy.deepcopy(bidsmap['DICOM']['func'][0])
    run['datasource'].datatype = 'anat'

    # Update the bidsmap
    bids.update_bidsmap(bidsmap, 'func', run)
    assert bidsmap['DICOM']['anat'][-1]['provenance'] == run['provenance']
    assert bidsmap['DICOM']['func'] [0]['provenance'] != run['provenance']

    # Modify the last anat run-item and update the bidsmap
    run['bids']['foo'] = 'bar'
    bids.update_bidsmap(bidsmap, 'anat', run)
    assert bidsmap['DICOM']['anat'][-1]['bids']['foo'] == 'bar'


def test_exist_run(test_bidsmap):

    # Load a bidsmap
    bidsmap, _ = bids.load_bidsmap(test_bidsmap)

    # Collect the first anat run-item
    run = copy.deepcopy(bidsmap['DICOM']['anat'][0])

    # Find the run in the wrong datatype
    assert bids.exist_run(bidsmap, 'func', run) == False

    # Find run with in the right datatype and in all datatypes
    assert bids.exist_run(bidsmap, 'anat', run) == True
    assert bids.exist_run(bidsmap, '',     run) == True

    # Find the wrong run in all datatypes
    run['attributes']['ProtocolName'] = 'abcdefg'
    assert bids.exist_run(bidsmap, '', run)     == False
    run['attributes']['ProtocolName'] = ''
    assert bids.exist_run(bidsmap, '', run)     == False


def test_increment_runindex_no_run1(tmp_path):
    """Test if run-index is preserved or added to the bidsname"""

    # Test runindex is <<>>, so no run is added to the bidsname
    outfolder = tmp_path/'bids'/'sub-01'/'anat'
    bidsname  = bids.increment_runindex(outfolder, 'sub-01_T1w', {'bids': {'run': '<<>>'}})
    assert bidsname == 'sub-01_T1w'

    # Test runindex is <<1>>, so run-1 is preserved in the bidsname
    bidsname = bids.increment_runindex(outfolder, 'sub-01_run-1_T1w', {'bids': {'run': '<<1>>'}})
    assert bidsname == 'sub-01_run-1_T1w'

    # Test runindex is <<2>>, so run-2 is preserved in the bidsname
    bidsname = bids.increment_runindex(outfolder, 'sub-01_run-2_T1w', {'bids': {'run': '<<2>>'}})
    assert bidsname == 'sub-01_run-2_T1w'


def test_increment_runindex_rename_run1(tmp_path):
    """Test runindex is <<>>, so run-2 is added to the bidsname and existing run-less files are renamed to run-1"""

    # Create the run-less files
    old_run1name = 'sub-01_T1w'
    new_run1name = 'sub-01_run-1_T1w'
    outfolder    = tmp_path/'bids'/'sub-01'/'anat'
    outfolder.mkdir(parents=True)
    for suffix in ('.nii.gz', '.json'):
        (outfolder/old_run1name).with_suffix(suffix).touch()

    # Create the scans table
    scans_data         = {'filename': ['anat/sub-01_T2w.nii.gz', f"anat/{old_run1name}.nii.gz"], 'acq_time': ['acq1', 'acq2']}  # One matching run-less file
    result_scans_data  = {'filename': ['anat/sub-01_T2w.nii.gz', f"anat/{new_run1name}.nii.gz"], 'acq_time': ['acq1', 'acq2']}  # One matching run-1 file
    scans_table        = pd.DataFrame(scans_data).set_index('filename')
    result_scans_table = pd.DataFrame(result_scans_data).set_index('filename')

    # Increment the run-index
    bidsname = bids.increment_runindex(outfolder, 'sub-01_T1w', {'bids': {'run': '<<>>'}}, scans_table)

    # Check the results
    assert bidsname == 'sub-01_run-2_T1w'
    assert result_scans_table.equals(scans_table)
    for suffix in ('.nii.gz', '.json'):
        assert (outfolder/old_run1name).with_suffix(suffix).is_file() == False
        assert (outfolder/new_run1name).with_suffix(suffix).is_file() == True


def test_increment_runindex_run1_run2_exists(tmp_path):
    """Test if run-3 is added to the bidsname"""

    # Create the run-1 and run-2 files
    outfolder = tmp_path/'bids'/'sub-01'/'anat'
    outfolder.mkdir(parents=True)
    for suffix in ('.nii.gz', '.json'):
        (outfolder/'sub-01_run-1_T1w').with_suffix(suffix).touch()
        (outfolder/'sub-01_run-2_T1w').with_suffix(suffix).touch()

    # Test run-index is <<>>, so the run-index is incremented
    bidsname = bids.increment_runindex(outfolder, 'sub-01_T1w.nii.gz', {'bids': {'run': '<<>>'}})
    assert bidsname == 'sub-01_run-3_T1w.nii.gz'

    # Test run-index is <<1>>, so the run-index is incremented
    bidsname = bids.increment_runindex(outfolder, 'sub-01_run-1_T1w', {'bids': {'run': '<<1>>'}})
    assert bidsname == 'sub-01_run-3_T1w'

    # Test run-index is <<AttrKey>>, so the run-index is untouched
    bidsname  = bids.increment_runindex(outfolder, 'sub-01_run-1_T1w', {'bids': {'run': '<<AttrKey>>'}})
    assert bidsname == 'sub-01_run-1_T1w'

    # Test run-index is 2, so the run-index is untouched
    bidsname  = bids.increment_runindex(outfolder, 'sub-01_run-1_T1w', {'bids': {'run': '2'}})
    assert bidsname == 'sub-01_run-1_T1w'


def test_get_bidsname(raw_dicomdir):

    dicomfile   = raw_dicomdir/'Doe^Archibald'/'01-XR C Spine Comp Min 4 Views'/'001-Cervical LAT'/'6154'
    run         = {'datasource': bids.DataSource(dicomfile, {'dcm2niix2bids': {}}, 'DICOM')}
    run['bids'] = {'acq':'py#dicom', 'foo@':'bar#123', 'run':'<<SeriesNumber>>', 'suffix':'T0w'}

    bidsname = bids.get_bidsname('sub-001', 'ses-01', run, validkeys=False, cleanup=False)  # Test default: runtime=False
    assert bidsname == 'sub-001_ses-01_acq-py#dicom_run-<<SeriesNumber>>_foo@-bar#123_T0w'

    bidsname = bids.get_bidsname('sub-001', 'ses-01', run, validkeys=False, runtime=False, cleanup=True)
    assert bidsname == 'sub-001_ses-01_acq-pydicom_run-SeriesNumber_foo@-bar123_T0w'

    bidsname = bids.get_bidsname('sub-001', 'ses-01', run, validkeys=False, runtime=True,  cleanup=False)
    assert bidsname == 'sub-001_ses-01_acq-py#dicom_run-1_foo@-bar#123_T0w'

    bidsname = bids.get_bidsname('sub-001', 'ses-01', run, validkeys=True,  runtime=True,  cleanup=False)
    assert bidsname == 'sub-001_ses-01_acq-py#dicom_run-1_T0w'

    run['bids']['run'] = '<<1>>'
    bidsname = bids.get_bidsname('sub-001', '', run, validkeys=True, runtime=True)          # Test default: cleanup=True
    assert bidsname == 'sub-001_acq-pydicom_run-1_T0w'

    run['bids']['run'] = '<<>>'
    bidsname = bids.get_bidsname('sub-001', '', run, validkeys=True, runtime=True)          # Test default: cleanup=True
    assert bidsname == 'sub-001_acq-pydicom_T0w'



from setuptools import setup
from build_manpages import build_manpages, get_build_py_cmd, get_install_cmd

setup(cmdclass = {'build_manpages': build_manpages,
                  'build_py': get_build_py_cmd(),
                  'install': get_install_cmd()})



#!/usr/bin/env python3
"""A BIDScoin application to convert source data to BIDS (See also cli/_bidscoiner.py)"""

import dateutil.parser
import re
import pandas as pd
import json
import logging
import shutil
import urllib.request
from tqdm import tqdm
from tqdm.contrib.logging import logging_redirect_tqdm
from pathlib import Path
from importlib.util import find_spec
if find_spec('bidscoin') is None:
    import sys
    sys.path.append(str(Path(__file__).parents[1]))
from bidscoin import bcoin, bids, bidsversion, __version__


def bidscoiner(rawfolder: str, bidsfolder: str, subjects: list=(), force: bool=False, bidsmapfile: str='bidsmap.yaml') -> None:
    """
    Main function that processes all the subjects and session in the sourcefolder and uses the
    bidsmap.yaml file in bidsfolder/code/bidscoin to cast the data into the BIDS folder.

    :param rawfolder:       The root folder-name of the sub/ses/data/file tree containing the source data files
    :param bidsfolder:      The name of the BIDS root folder
    :param subjects:        List of selected subjects / participants (i.e. sub-# names / folders) to be processed (the sub-prefix can be removed). Otherwise, all subjects in the sourcefolder will be selected
    :param force:           If True, subjects will be processed, regardless of existing folders in the bidsfolder. Otherwise, existing folders will be skipped
    :param bidsmapfile:     The name of the bidsmap YAML-file. If the bidsmap pathname is relative (i.e. no "/" in the name) then it is assumed to be located in bidsfolder/code/bidscoin
    :return:                Nothing
    """

    # Input checking & defaults
    rawfolder   = Path(rawfolder).resolve()
    bidsfolder  = Path(bidsfolder).resolve()
    bidsmapfile = Path(bidsmapfile)

    # Start logging
    bcoin.setup_logging(bidsfolder/'code'/'bidscoin'/'bidscoiner.log')
    LOGGER.info('')
    LOGGER.info(f"-------------- START BIDScoiner {__version__}: BIDS {bidsversion()} ------------")
    LOGGER.info(f">>> bidscoiner sourcefolder={rawfolder} bidsfolder={bidsfolder} subjects={subjects} force={force} bidsmap={bidsmapfile}")

    # Create a code/bidscoin subfolder
    (bidsfolder/'code'/'bidscoin').mkdir(parents=True, exist_ok=True)

    # Create a dataset description file if it does not exist
    dataset_file = bidsfolder/'dataset_description.json'
    generatedby  = [{"Name":"BIDScoin", "Version":__version__, "CodeURL":"https://github.com/Donders-Institute/bidscoin"}]
    if not dataset_file.is_file():
        LOGGER.info(f"Creating dataset description file: {dataset_file}")
        dataset_description = {"Name":                  "REQUIRED. Name of the dataset",
                               "GeneratedBy":           generatedby,
                               "BIDSVersion":           str(bidsversion()),
                               "DatasetType":           "raw",
                               "License":               "RECOMMENDED. The license for the dataset. The use of license name abbreviations is RECOMMENDED for specifying a license. The corresponding full license text MAY be specified in an additional LICENSE file",
                               "Authors":               ["OPTIONAL. List of individuals who contributed to the creation/curation of the dataset"],
                               "Acknowledgements":      "OPTIONAL. Text acknowledging contributions of individuals or institutions beyond those listed in Authors or Funding",
                               "HowToAcknowledge":      "OPTIONAL. Instructions how researchers using this dataset should acknowledge the original authors. This field can also be used to define a publication that should be cited in publications that use the dataset",
                               "Funding":               ["OPTIONAL. List of sources of funding (grant numbers)"],
                               "EthicsApprovals":    	["OPTIONAL. List of ethics committee approvals of the research protocols and/or protocol identifiers"],
                               "ReferencesAndLinks":    ["OPTIONAL. List of references to publication that contain information on the dataset, or links", "https://github.com/Donders-Institute/bidscoin"],
                               "DatasetDOI":            "OPTIONAL. The Document Object Identifier of the dataset (not the corresponding paper)"}
    else:
        with dataset_file.open('r') as fid:
            dataset_description = json.load(fid)
        if 'BIDScoin' not in [generatedby_['Name'] for generatedby_ in dataset_description.get('GeneratedBy',[])]:
            LOGGER.verbose(f"Adding {generatedby} to {dataset_file}")
            dataset_description['GeneratedBy'] = dataset_description.get('GeneratedBy',[]) + generatedby
    with dataset_file.open('w') as fid:
        json.dump(dataset_description, fid, indent=4)

    # Create a README file if it does not exist
    readme_file = bidsfolder/'README'
    if not readme_file.is_file():
        LOGGER.info(f"Creating a template README file (adjust it to your needs): {readme_file}")
        try:
            urllib.request.urlretrieve('https://raw.githubusercontent.com/bids-standard/bids-starter-kit/main/templates/README.MD', readme_file)
        except urllib.error.URLError:
            readme_file.write_text(
                f"A free form text ( README ) describing the dataset in more details that SHOULD be provided. For an example, see e.g.:\n"
                f"https://github.com/bids-standard/bids-starter-kit/blob/main/templates/README.MD\n\n"
                f"The raw BIDS data was created using BIDScoin {__version__}\n"
                f"All provenance information and settings can be found in ./code/bidscoin\n"
                f"For more information see: https://github.com/Donders-Institute/bidscoin\n")

    # Get the bidsmap heuristics from the bidsmap YAML-file
    bidsmap, _  = bids.load_bidsmap(bidsmapfile, bidsfolder/'code'/'bidscoin')
    dataformats = [dataformat for dataformat in bidsmap if dataformat and dataformat != 'Options']
    if not bidsmap:
        LOGGER.error(f"No bidsmap file found in {bidsfolder}. Please run the bidsmapper first and/or use the correct bidsfolder")
        return

    # Load the data conversion plugins
    plugins = [bcoin.import_plugin(plugin, ('bidscoiner_plugin',)) for plugin,options in bidsmap['Options']['plugins'].items()]
    plugins = [plugin for plugin in plugins if plugin]          # Filter the empty items from the list
    if not plugins:
        LOGGER.warning(f"The plugins listed in your bidsmap['Options'] did not have a usable `bidscoiner_plugin` function, nothing to do")
        LOGGER.info('-------------- FINISHED! ------------')
        LOGGER.info('')
        return

    # Append options to the .bidsignore file
    bidsignore_items = [item.strip() for item in bidsmap['Options']['bidscoin']['bidsignore'].split(';')]
    bidsignore_file  = bidsfolder/'.bidsignore'
    if bidsignore_items:
        LOGGER.verbose(f"Writing {bidsignore_items} entries to {bidsignore_file}")
        if bidsignore_file.is_file():
            bidsignore_items += bidsignore_file.read_text().splitlines()
        with bidsignore_file.open('w') as bidsignore:
            for item in set(bidsignore_items):
                bidsignore.write(item + '\n')

    # Get the table & dictionary of the subjects that have been processed
    participants_tsv = bidsfolder/'participants.tsv'
    if participants_tsv.is_file():
        participants_table = pd.read_csv(participants_tsv, sep='\t')
        participants_table.set_index(['participant_id'], verify_integrity=True, inplace=True)
    else:
        participants_table = pd.DataFrame()
        participants_table.index.name = 'participant_id'

    # Get the list of subjects
    subprefix = bidsmap['Options']['bidscoin']['subprefix'].replace('*','')
    sesprefix = bidsmap['Options']['bidscoin']['sesprefix'].replace('*','')
    if not subjects:
        subjects = bcoin.lsdirs(rawfolder, (subprefix if subprefix!='*' else '') + '*')
        if not subjects:
            LOGGER.warning(f"No subjects found in: {rawfolder/subprefix}*")
    else:
        subjects = [rawfolder/(subprefix + re.sub(f"^{'' if subprefix=='*' else re.escape(subprefix)}",'',subject)) for subject in subjects]   # Make sure there is a sub-prefix

    # Loop over all subjects and sessions and convert them using the bidsmap entries
    with logging_redirect_tqdm():
        for n, subject in enumerate(tqdm(subjects, unit='subject', leave=False), 1):

            LOGGER.info(f"------------------- Subject {n}/{len(subjects)} -------------------")
            if not subject.is_dir():
                LOGGER.error(f"The '{subject}' subject folder does not exist")
                continue

            sessions = bcoin.lsdirs(subject, (sesprefix if sesprefix!='*' else '') + '*')
            if not sessions or (subject/'DICOMDIR').is_file():
                sessions = [subject]
            for session in sessions:

                # Unpack the data in a temporary folder if it is tarballed/zipped and/or contains a DICOMDIR file
                sesfolders, unpacked = bids.unpack(session, bidsmap['Options']['bidscoin'].get('unzip',''))
                for sesfolder in sesfolders:

                    # Check if we should skip the session-folder
                    datasource = bids.get_datasource(sesfolder, bidsmap['Options']['plugins'])
                    if not datasource.dataformat:
                        LOGGER.info(f">>> No coinable datasources found in '{sesfolder}'")
                        continue
                    subid        = bidsmap[datasource.dataformat]['subject']
                    sesid        = bidsmap[datasource.dataformat]['session']
                    subid, sesid = datasource.subid_sesid(subid, sesid if sesid else '')
                    bidssession  = bidsfolder/subid/sesid       # TODO: Support DICOMDIR with multiple subjects (as in PYDICOMDIR)
                    if not force and bidssession.is_dir():
                        datatypes = []
                        for dataformat in dataformats:
                            for datatype in bcoin.lsdirs(bidssession):                               # See what datatypes we already have in the bids session-folder
                                if list(datatype.iterdir()) and bidsmap[dataformat].get(datatype.name): # See if we are going to add data for this datatype
                                    datatypes.append(datatype.name)
                        if datatypes:
                            LOGGER.info(f">>> Skipping processed session: {bidssession} already has {datatypes} data (you can carefully use the -f option to overrule)")
                            continue

                    LOGGER.info(f">>> Coining datasources in: {sesfolder}")
                    if bidssession.is_dir():
                        LOGGER.warning(f"Existing BIDS output-directory found, which may result in duplicate data (with increased run-index). Make sure {bidssession} was cleaned-up from old data before (re)running the bidscoiner")
                    bidssession.mkdir(parents=True, exist_ok=True)

                    # Run the bidscoiner plugins
                    for module in plugins:
                        LOGGER.verbose(f"Executing plugin: {Path(module.__file__).name}")
                        module.bidscoiner_plugin(sesfolder, bidsmap, bidssession)

                    # Add the special fieldmap metadata (IntendedFor, B0FieldIdentifier, TE, etc)
                    addmetadata(bidssession, subid, sesid)

                    # Clean-up the temporary unpacked data
                    if unpacked:
                        shutil.rmtree(sesfolder)

    # Re-read the participants_table (the plugins may have changed it) and store the collected personals in the json sidecar-file
    if participants_tsv.is_file():
        participants_table = pd.read_csv(participants_tsv, sep='\t')
        participants_table.set_index(['participant_id'], verify_integrity=True, inplace=True)
    participants_json = participants_tsv.with_suffix('.json')
    participants_dict = {}
    if participants_json.is_file():
        with participants_json.open('r') as json_fid:
            participants_dict = json.load(json_fid)
    if not participants_dict.get('participant_id'):
        participants_dict['participant_id'] = {'Description': 'Unique participant identifier'}
    if not participants_dict.get('session_id') and 'session_id' in participants_table.columns:
        participants_dict['session_id'] = {'Description': 'Session identifier'}
    newkey = False
    for col in participants_table.columns:
        if col not in participants_dict:
            newkey = True
            participants_dict[col] = dict(LongName    = 'Long (unabbreviated) name of the column',
                                          Description = 'Description of the the column',
                                          Levels      = dict(Key='Value (This is for categorical variables: a dictionary of possible values (keys) and their descriptions (values))'),
                                          Units       = 'Measurement units. [<prefix symbol>]<unit symbol> format following the SI standard is RECOMMENDED')

    # Write the collected data to the participant files
    if newkey:
        LOGGER.info(f"Writing subject meta data to: {participants_json}")
        with participants_json.open('w') as json_fid:
            json.dump(participants_dict, json_fid, indent=4)

    LOGGER.info('-------------- FINISHED! ------------')
    LOGGER.info('')

    bcoin.reporterrors()


def addmetadata(bidsses: Path, subid: str, sesid: str) -> None:
    """
    Adds the special fieldmap metadata (IntendedFor, B0FieldIdentifier, TE, etc.)

    :param bidsses: The session folder with the BIDS session data
    :param subid:   The subject 'sub-label' identifier
    :param sesid:   The session 'ses-label' identifier
    """

    # Add IntendedFor search results and TE1+TE2 meta-data to the fieldmap json-files. This has been postponed until all datatypes have been processed (i.e. so that all target images are indeed on disk)
    if (bidsses/'fmap').is_dir():

        scans_tsv = bidsses/f"{subid}{'_'+sesid if sesid else ''}_scans.tsv"
        if scans_tsv.is_file():
            scans_table = pd.read_csv(scans_tsv, sep='\t', index_col='filename')
        else:
            scans_table = pd.DataFrame(columns=['acq_time'])

        fmaps = [fmap.relative_to(bidsses).as_posix() for fmap in sorted((bidsses/'fmap').glob('sub-*.nii*'))]
        for fmap in fmaps:

            # Load the existing meta-data
            jsonfile = bidsses/Path(fmap).with_suffix('').with_suffix('.json')
            with jsonfile.open('r') as sidecar:
                jsondata = json.load(sidecar)

            # Search for the imaging files that match the IntendedFor search criteria
            intendedfor = jsondata.get('IntendedFor')
            if intendedfor and isinstance(intendedfor, str):

                # Check if there are multiple runs and get the lower- and upperbound from the AcquisitionTime to limit down the IntendedFor search
                fmaptime   = dateutil.parser.parse('1925-01-01')                                    # If nothing, use the BIDS stub acquisition time
                lowerbound = fmaptime.replace(year=1900)                                            # If nothing, use an ultra-wide lower limit for the IntendedFor search
                upperbound = fmaptime.replace(year=2100)                                            # Idem for the upper limit
                try:                                                                                # There may be more fieldmaps, hence try to limit down the search to the adjacently acquired data
                    fmaptime = dateutil.parser.parse(scans_table.loc[fmap, 'acq_time'])
                    runindex = bids.get_bidsvalue(fmap, 'run')
                    prevfmap = bids.get_bidsvalue(fmap, 'run', str(int(runindex) - 1))
                    nextfmap = bids.get_bidsvalue(fmap, 'run', str(int(runindex) + 1))
                    if prevfmap in fmaps:
                        lowerbound = dateutil.parser.parse(scans_table.loc[prevfmap, 'acq_time'])   # Narrow the lower search limit down to the preceding fieldmap
                    if nextfmap in fmaps:
                        upperbound = dateutil.parser.parse(scans_table.loc[nextfmap, 'acq_time'])   # Narrow the upper search limit down to the succeeding fieldmap
                except (TypeError, ValueError, KeyError, dateutil.parser.ParserError) as acqtimeerror:
                    pass                                                                            # Raise this only if there are limits and matches, i.e. below

                # Search with multiple patterns for matching NIfTI-files in all runs and store the relative path to the session folder
                niifiles = []
                if intendedfor.startswith('<') and intendedfor.endswith('>'):
                    intendedfor = intendedfor[2:-2].split('><')
                elif not isinstance(intendedfor, list):
                    intendedfor = [intendedfor]
                for part in intendedfor:
                    limits  = part.split(':',1)[1].strip() if ':' in part else ''   # part = 'pattern: [lowerlimit:upperlimit]'
                    pattern = part.split(':',1)[0].strip()
                    matches = [niifile.relative_to(bidsses).as_posix() for niifile in sorted(bidsses.rglob(f"*{pattern}*")) if pattern and '.nii' in niifile.suffixes]
                    if limits and matches:
                        try:
                            limits     = limits[1:-1].split(':',1)                  # limits: '[lowerlimit:upperlimit]' -> ['lowerlimit', 'upperlimit']
                            lowerlimit = int(limits[0]) if limits[0].strip() else float('-inf')
                            upperlimit = int(limits[1]) if limits[1].strip() else float('inf')
                            acqtimes   = []
                            for match in matches:
                                acqtimes.append((dateutil.parser.parse(scans_table.loc[match,'acq_time']), match))      # Time + filepath relative to the session-folder
                            acqtimes.sort(key = lambda acqtime: acqtime[0])
                            offset = sum([acqtime[0] < fmaptime for acqtime in acqtimes])  # The nr of preceding series
                            for n, acqtime in enumerate(acqtimes):
                                if lowerbound < acqtime[0] < upperbound and lowerlimit <= n-offset < upperlimit:
                                    niifiles.append(acqtime[1])
                        except Exception as intendedforerror:
                            LOGGER.error(f"Could not bound the <{part}> IntendedFor search as it requires a *_scans.tsv file with acq_time values for: {fmap}\n{intendedforerror}")
                            niifiles.extend(matches)
                    else:
                        niifiles.extend(matches)

                # Add the IntendedFor data. NB: The BIDS URI paths need to use forward slashes and be relative to the bids root folder
                if niifiles:
                    LOGGER.verbose(f"Adding IntendedFor to: {jsonfile}")
                    jsondata['IntendedFor'] = [f"bids::{(Path(subid)/sesid/niifile).as_posix()}" for niifile in niifiles]
                else:
                    LOGGER.warning(f"Empty 'IntendedFor' fieldmap value in {jsonfile}: the search for {intendedfor} gave no results")
                    jsondata['IntendedFor'] = None

            elif not (jsondata.get('B0FieldSource') or jsondata.get('B0FieldIdentifier')):
                LOGGER.warning(f"Empty IntendedFor / B0FieldSource / B0FieldIdentifier fieldmap values in {jsonfile} (i.e. the fieldmap may not be used)")

            # Work-around because the bids-validator (v1.8) cannot handle `null` values / unused IntendedFor fields
            if not jsondata.get('IntendedFor'):
                jsondata.pop('IntendedFor', None)

            # Extract the echo times from magnitude1 and magnitude2 and add them to the phasediff json-file
            if jsonfile.name.endswith('phasediff.json'):
                json_magnitude = [None, None]
                echotime       = [None, None]
                for n in (0,1):
                    json_magnitude[n] = jsonfile.parent/jsonfile.name.replace('_phasediff', f"_magnitude{n+1}")
                    if not json_magnitude[n].is_file():
                        LOGGER.error(f"Could not find expected magnitude{n+1} image associated with: {jsonfile}\nUse the bidseditor to verify that the fmap images that belong together have corresponding BIDS output names")
                    else:
                        with json_magnitude[n].open('r') as sidecar:
                            data = json.load(sidecar)
                        echotime[n] = data.get('EchoTime')
                jsondata['EchoTime1'] = jsondata['EchoTime2'] = None
                if None in echotime:
                    LOGGER.error(f"Cannot find and add valid EchoTime1={echotime[0]} and EchoTime2={echotime[1]} data to: {jsonfile}")
                elif echotime[0] > echotime[1]:
                    LOGGER.error(f"Found invalid EchoTime1={echotime[0]} > EchoTime2={echotime[1]} for: {jsonfile}")
                else:
                    jsondata['EchoTime1'] = echotime[0]
                    jsondata['EchoTime2'] = echotime[1]
                    LOGGER.verbose(f"Adding EchoTime1: {echotime[0]} and EchoTime2: {echotime[1]} to {jsonfile}")

            # Save the collected meta-data to disk
            with jsonfile.open('w') as sidecar:
                json.dump(jsondata, sidecar, indent=4)


def main():
    """Console script entry point"""

    from bidscoin.cli._bidscoiner import get_parser

    # Parse the input arguments and run bidscoiner(args)
    args = get_parser().parse_args()
    bidscoiner(rawfolder   = args.sourcefolder,
               bidsfolder  = args.bidsfolder,
               subjects    = args.participant_label,
               force       = args.force,
               bidsmapfile = args.bidsmap)


if __name__ == "__main__":
    LOGGER = logging.getLogger(f"bidscoin.{Path(__file__).stem}")
    main()

else:
    LOGGER = logging.getLogger(__name__)



#!/usr/bin/env python3
"""A bidsapp that combines echos and wraps around FreeSurfer's 'synthstrip' skull stripping tool (See also cli/_skullstrip.py)"""

import shutil
import json
import logging
import pandas as pd
import nibabel as nib
from tqdm import tqdm
from tqdm.contrib.logging import logging_redirect_tqdm
from pathlib import Path
from importlib.util import find_spec
if find_spec('bidscoin') is None:
    import sys
    sys.path.append(str(Path(__file__).parents[2]))
from bidscoin import bcoin, bids


def skullstrip(bidsdir: str, pattern: str, subjects: list, masked: str, output: list, force: bool, args: str, cluster: bool):
    """
    :param bidsdir:     The bids-directory with the subject data
    :param pattern:     Globlike search pattern (relative to the subject/session folder) to select the images that need to be skullstripped, e.g. 'anat/*_T1w*'
    :param subjects:    List of sub-# identifiers to be processed (the sub-prefix can be left out). If not specified then all sub-folders in the bidsfolder will be processed
    :param force:       If True then images will be processed, regardless if images have already been skullstripped (i.e. if {"SkullStripped": True} in the json sidecar file)
    :param masked:      Globlike search pattern (relative to the subject/session folder) to select additional images that need to be masked with the same mask, e.g. 'fmap/*_phasediff')
    :param output:      One or two output strings that determine where the skullstripped + additional masked images are saved. Each output string can be the name of a BIDS datatype folder, such as 'anat', or of the derivatives folder, i.e. 'derivatives' (default). If the output string is the same as the datatype then the original images are replaced by the skullstripped images
    :param args:        Additional arguments that are passed to synthstrip. See examples for usage
    :param cluster:     Use qsub to submit the skullstrip jobs to a high-performance compute (HPC) cluster
    :return:
    """

    # Input checking
    bidsdir = Path(bidsdir).resolve()
    if not bidsdir.is_dir():
        print(f"Could not find the bids folder: {bidsdir}")
        return
    if not output or len(output) == 0:
        output = ['derivatives']
    if len(output) == 1:
        output += output
    if len(output) != 2:
        print(f"The 'output' argument should be one or strings, not {len(output)} ({output})")
        return
    if not shutil.which('mri_synthstrip'):
        print("Could not find 'mri_synthstrip', skullstrip requires FreeSurfer v7.3.2 or higher")
        return
    if cluster and masked:
        print("The `--cluster` option cannot be used in combination with `--masked`")
        return

    # Start logging
    bcoin.setup_logging(bidsdir/'code'/'bidscoin'/'skullstrip.log')
    LOGGER.info('')
    LOGGER.info('------------ START skullstrip ------------')
    LOGGER.info(f">>> skullstrip bidsfolder={bidsdir} pattern={pattern} subjects={subjects} masked={masked} output={output} force={force} {args}")

    # Get the list of subjects
    if not subjects:
        subjects = bcoin.lsdirs(bidsdir, 'sub-*')
        if not subjects:
            LOGGER.warning(f"No subjects found in: {bidsdir/'sub-*'}")
    else:
        subjects = ['sub-' + subject.replace('sub-', '') for subject in subjects]               # Make sure there is a "sub-" prefix
        subjects = [bidsdir/subject for subject in subjects if (bidsdir/subject).is_dir()]

    # Loop over bids subject/session-directories
    with logging_redirect_tqdm():
        for n, subject in enumerate(tqdm(subjects, unit='subject', leave=False), 1):

            subid    = subject.name
            sessions = bcoin.lsdirs(subject, 'ses-*')
            if not sessions:
                sessions = [subject]
            for session in sessions:

                LOGGER.info('--------------------------------------')
                LOGGER.info(f"Processing ({n}/{len(subjects)}): {session}")

                # Search for images that need to be skullstripped
                sesid   = session.name if session.name.startswith('ses-') else ''
                srcimgs = sorted([match for match in session.glob(pattern) if '.nii' in match.suffixes])
                addimgs = sorted([match for match in session.glob(masked)  if '.nii' in match.suffixes]) if masked else []
                if not srcimgs:
                    LOGGER.info(f"No images found for {pattern} in: {session}")
                if masked and not addimgs:
                    LOGGER.info(f"No images found for {masked} in: {session}")
                if addimgs and len(srcimgs) > 1:
                    LOGGER.error(f"{len(srcimgs)} matches found for {session/pattern}, which is ambiguous (the {masked} masked option requires a single match)")
                    addimgs = []
                for srcimg in srcimgs:

                    # Construct the output filename and relative path names (for updating the IntendedFor list and scans.tsv file)
                    srcent, suffix = srcimg.with_suffix('').stem.rsplit('_',1)  # Name without suffix, suffix
                    derent         = 'space-orig_desc-brain'
                    ext            = ''.join(srcimg.suffixes)                   # Account for e.g. '.nii.gz'
                    if output[0] == 'derivatives':
                        outputimg = bidsdir/'derivatives'/'skullstrip'/subid/sesid/srcimg.parent.name/f"{srcent}_{derent}_{suffix}{ext}"
                    else:
                        outputimg = session/output[0]/srcimg.name               # NB: Overwrite the original image
                    outputimg.parent.mkdir(parents=True, exist_ok=True)

                    # Check the json "SkullStripped" field to see if it has already been skullstripped
                    outputjson = outputimg.with_suffix('').with_suffix('.json')
                    if not force and outputjson.is_file():
                        with outputjson.open('r') as sidecar:
                            metadata = json.load(sidecar)
                        if metadata.get('SkullStripped'):
                            LOGGER.info(f"Skipping already skullstripped image: {outputimg}")
                            continue

                    # Skullstrip the image
                    maskimg = bidsdir/'derivatives'/'skullstrip'/subid/sesid/srcimg.parent.name/f"{srcent}_{derent}_mask{ext}"
                    maskimg.parent.mkdir(parents=True, exist_ok=True)
                    command = f"mri_synthstrip -i {srcimg} -o {outputimg} -m {maskimg} {args}"
                    if cluster:
                        command = f"qsub -l walltime=0:05:00,mem=8gb -N skullstrip_{subid}_{sesid} <<EOF\n{command}\nEOF"
                    if bcoin.run_command(command):
                        continue

                    # Add a json sidecar-file with the "SkullStripped" field
                    srcjson = srcimg.with_suffix('').with_suffix('.json')
                    if srcjson.is_file():
                        with srcjson.open('r') as sidecar:
                            metadata = json.load(sidecar)
                    else:
                        metadata = {}
                    metadata['SkullStripped'] = True
                    with outputjson.open('w') as sidecar:
                        json.dump(metadata, sidecar, indent=4)
                    metadata['Type'] = 'Brain'
                    with maskimg.with_suffix('').with_suffix('.json').open('w') as sidecar:
                        json.dump(metadata, sidecar, indent=4)

                    # Apply the mask to the additional images and save it as output
                    maskobj    = nib.load(maskimg)   if addimgs else []
                    maskvol    = maskobj.get_fdata() if addimgs else []
                    addoutimgs = []                # To be used for updating the IntendedFor list and the scans.tsv file
                    for addimg in addimgs:

                        # Define the output image
                        if output[1] == 'derivatives':
                            srcent, suffix = addimg.name.rsplit('_', 1)     # NB: suffix = suffix.ext
                            addoutimg      = bidsdir/'derivatives'/'skullstrip'/subid/sesid/addimg.parent.name/f"{srcent}_{derent}_{suffix}"
                        else:
                            addoutimg = session/output[1]/addimg.name       # NB: Could be the same as the original image
                        addoutimgs.append(addoutimg)
                        outputjson = addoutimg.with_suffix('').with_suffix('.json')
                        if not force and outputjson.is_file():
                            with outputjson.open('r') as sidecar:
                                metadata = json.load(sidecar)
                            if metadata.get('SkullStripped'):
                                LOGGER.info(f"Skipping already skullstripped image: {addoutimg}")
                                continue

                        # Load the volume data, multiply it with the mask and save it to the output image
                        addoutimg.parent.mkdir(parents=True, exist_ok=True)
                        addobj = nib.load(addimg)
                        if addobj.header.get_data_shape()[0:3] == maskobj.header.get_data_shape():
                            LOGGER.info(f"Applying skullstrip-mask to: {addoutimg}")
                            addmsk = nib.Nifti1Image(addobj.get_fdata() * maskvol, addobj.affine, addobj.header)
                            addmsk.to_filename(addoutimg)
                        else:
                            LOGGER.error(f"Cannot apply skullstrip-mask to: {addoutimg}\nIt's dimensions are {addobj.header.get_data_shape()} but the mask is {maskobj.header.get_data_shape()}")
                            continue

                        # Add a json sidecar-file to the output image
                        if addimg.with_suffix('').with_suffix('.json').is_file():
                            with addimg.with_suffix('').with_suffix('.json').open('r') as sidecar:
                                metadata = json.load(sidecar)
                        else:
                            metadata = {}
                        metadata['SkullStripped'] = True
                        with outputjson.open('w') as sidecar:
                            json.dump(metadata, sidecar, indent=4)

                    # Update the scans.tsv file
                    scans_tsv = session/f"{subid}{'_'+sesid if sesid else ''}_scans.tsv"
                    if scans_tsv.is_file():

                        # Add the new images to the scans table
                        scans_table = pd.read_csv(scans_tsv, sep='\t', index_col='filename')
                        bidsignore  = (bidsdir/'.bidsignore').read_text().splitlines() if (bidsdir/'.bidsignore').is_file() else ['extra_data/']
                        for original, stripped in zip([srcimg] + addimgs, [outputimg] + addoutimgs):
                            if not stripped.name or 'derivatives' in stripped.parts or bids.check_ignore(stripped.parent.name, bidsignore):
                                continue
                            original_rel = original.relative_to(session).as_posix()
                            stripped_rel = stripped.relative_to(session).as_posix()
                            datatype     = stripped_rel.split('/')[0]
                            if original_rel in scans_table.index and stripped_rel not in scans_table.index and datatype in bids.datatyperules:
                                LOGGER.info(f"Adding '{stripped_rel}' to '{scans_tsv}'")
                                scans_table.loc[stripped_rel] = scans_table.loc[original_rel]

                        # Save the data
                        scans_table.sort_values(by=['acq_time','filename'], inplace=True)
                        scans_table.replace('','n/a').to_csv(scans_tsv, sep='\t', encoding='utf-8', na_rep='n/a')
                        for scan in scans_table.index:
                            if not (session/scan).is_file():
                                LOGGER.warning(f"Found non-existent file '{scan}' in '{scans_tsv}'")

    LOGGER.info('-------------- FINISHED! -------------')
    LOGGER.info('')
    if cluster:
        LOGGER.info('But first wait for your `skullstrip`-jobs to finish... Use e.g.:\n\nqstat $(qselect -s RQ) | grep skullstrip\n')


def main():
    """Console script entry point"""

    from bidscoin.cli._skullstrip import get_parser

    args = get_parser().parse_args()
    skullstrip(bidsdir  = args.bidsfolder,
               pattern  = args.pattern,
               subjects = args.participant_label,
               masked   = args.masked,
               output   = args.output,
               force    = args.force,
               args     = args.args,
               cluster  = args.cluster)


if __name__ == '__main__':
    LOGGER = logging.getLogger(f"bidscoin.{Path(__file__).stem}")
    main()

else:
    LOGGER = logging.getLogger(__name__)



import shutil
import csv
from pydicom.data import get_testdata_file
from pathlib import Path
from bidscoin import bcoin
from bidscoin.utilities import dicomsort, rawmapper, bidsparticipants

bcoin.setup_logging()


def test_dicomsort(tmp_path):
    shutil.copytree(Path(get_testdata_file('DICOMDIR')).parent, tmp_path, dirs_exist_ok=True)
    session = dicomsort.sortsessions(tmp_path/'DICOMDIR', folderscheme='{SeriesNumber:04d}-{SeriesDescription}', namescheme='{SeriesNumber:02d}_{SeriesDescription}_{AcquisitionNumber}_{InstanceNumber}.IMA', force=True)
    assert dicomsort.sortsessions(tmp_path/'DICOMDIR', folderscheme='{SeriesNumber:04d]-{SeriesDescription}') == [] # Invalid scheme -> clean return
    assert dicomsort.validscheme('{foo:04d}-{123}') == True
    assert dicomsort.validscheme('{foo:04d]-{bar}') == False
    assert dicomsort.validscheme('{foo:04}-{bar}')  == False
    assert (tmp_path/'Doe^Peter').is_dir()                                                                          # Subject (Patient): 98890234 -> Doe^Peter
    assert (tmp_path/'Doe^Archibald').is_dir()                                                                      #                    77654033 -> Doe^Archibald
    assert len(list((tmp_path/'Doe^Archibald').rglob('*')))                 == 13                                   #  6 directories +  7 files
    assert len(list((tmp_path/'Doe^Peter').rglob('*')))                     == 37                                   # 13 directories + 24 files
    assert session[0].parent.name                                           == 'Doe^Archibald'                      # Subject (Patient)
    assert session[0].name                                                  == '01-XR C Spine Comp Min 4 Views'     # Session (Study)
    assert sorted(sorted(session[0].iterdir())[0].iterdir())[0].parent.name == '0001-Cervical LAT'                  # Series -> folderscheme
    assert sorted(sorted(session[0].iterdir())[0].iterdir())[0].name        == '01_Cervical LAT__1.IMA'             # File   -> namescheme


def test_bidsparticipants(raw_dicomdir, bids_dicomdir, bidsmap_dicomdir):
    participantsfile = bids_dicomdir/'participants.tsv'
    if participantsfile.is_file():                                  # Compare newly generated data with reference data from test_bidscoiner
        with open(participantsfile) as fid:
            refdata = list(csv.reader(fid, delimiter='\t'))
        bidsparticipants.bidsparticipants(raw_dicomdir, bids_dicomdir, ['age', 'sex', 'size', 'weight'], bidsmapfile=bidsmap_dicomdir)
        with open(participantsfile) as fid:
            newdata = list(csv.reader(fid, delimiter='\t'))
        assert newdata == refdata


def test_rawmapper(raw_dicomdir, tmp_path):
    shutil.copytree(raw_dicomdir, tmp_path, dirs_exist_ok=True)

    mapperfile = tmp_path/'rawmapper_PatientID_PatientComments.tsv'
    rawmapper.rawmapper(tmp_path, dicomfield=('PatientID','PatientComments'), subprefix='Doe^', sesprefix='*')
    with open(mapperfile) as fid:
        mapperdata = list(csv.reader(fid, delimiter='\t'))
    assert len(mapperdata) == 7
    assert mapperdata[0]   == ['subid', 'sesid', 'seriesname', 'PatientID', 'PatientComments']
    assert mapperdata[1]   == ['Doe^Archibald', '01-XR C Spine Comp Min 4 Views', '001-Cervical LAT', '77654033', '']
    assert mapperdata[2]   == ['Doe^Archibald', '02-CT, HEADBRAIN WO CONTRAST', '002-Routine Brain', '77654033', '']

    mapperfile = tmp_path/'rawmapper_PatientID.tsv'
    rawmapper.rawmapper(tmp_path, dicomfield=('PatientID',), rename=True, subprefix='Doe^', sesprefix='01')
    with open(mapperfile) as fid:
        mapperdata = list(csv.reader(fid, delimiter='\t'))
    assert len(mapperdata) == 3
    assert mapperdata[0]   == ['subid', 'sesid', 'newsubid', 'newsesid']
    assert mapperdata[1]   == ['Doe^Archibald', '01-XR C Spine Comp Min 4 Views', 'Doe^77654033', '01-XR C Spine Comp Min 4 Views']
    assert mapperdata[2]   == ['Doe^Peter', '01-', 'Doe^98890234', '01-']
    assert (tmp_path/'Doe^77654033'/'01-XR C Spine Comp Min 4 Views').is_dir()
    assert (tmp_path/'Doe^98890234'/'01-').is_dir()

    rawmapper.rawmapper(tmp_path, dicomfield=('PatientID',), rename=True, subprefix='Doe^', sesprefix='02')
    with open(mapperfile) as fid:
        mapperdata = list(csv.reader(fid, delimiter='\t'))
    assert len(mapperdata) == 5
    assert mapperdata[0]   == ['subid', 'sesid', 'newsubid', 'newsesid']
    assert mapperdata[1]   == ['Doe^Archibald', '01-XR C Spine Comp Min 4 Views', 'Doe^77654033', '01-XR C Spine Comp Min 4 Views']
    assert mapperdata[4]   == ['Doe^Peter', '02-Carotids', 'Doe^98890234', '02-Carotids']
    assert (tmp_path/'Doe^98890234'/'02-Carotids').is_dir()

    rawmapper.rawmapper(tmp_path, dicomfield=('PatientID',), rename=True, subprefix='Doe^', sesprefix='*')
    with open(mapperfile) as fid:
        mapperdata = list(csv.reader(fid, delimiter='\t'))
    assert len(mapperdata) == 7
    assert mapperdata[6]   == ['Doe^Peter', '04-Brain-MRA', 'Doe^98890234', '04-Brain-MRA']
    assert (tmp_path/'Doe^98890234'/'04-Brain-MRA').is_dir()
